package cs108;

import java.io.IOException;
import java.io.InputStream;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

public final class LZWInputStream extends InputStream {
    private static final int DICT_CAPACITY = 4096;

    private final Bits12InputStream in;
    private final List<List<Integer>> dict;
    private final Deque<Integer> currList;
    private List<Integer> prevList;

    public LZWInputStream(InputStream in) {
        List<List<Integer>> dict = new ArrayList<>(DICT_CAPACITY);
        for (int i = 0; i <= 0xFF; i += 1)
            dict.add(List.of(i));

        this.in = new Bits12InputStream(in);
        this.dict = dict;
        this.currList = new ArrayDeque<>();
        this.prevList = null;
    }

    @Override
    public int read() throws IOException {
        if (currList.isEmpty()) {
            int i = in.readU12();
            if (i == -1) return -1;

            List<Integer> newList = i < dict.size()
                    ? dict.get(i)
                    : copyAndAppendHead(prevList, prevList);

            if (dict.size() < DICT_CAPACITY && prevList != null)
                dict.add(copyAndAppendHead(prevList, newList));

            currList.addAll(newList);
            prevList = newList;
        }
        return currList.removeFirst();
    }

    private static List<Integer> copyAndAppendHead(List<Integer> l, List<Integer> l2) {
        List<Integer> result = new ArrayList<>(l.size() + 1);
        result.addAll(l);
        result.add(l2.getFirst());
        return result;

        // Ou alors, en utilisant la programmation par flots :
//        return Stream.concat(l.stream(), l2.stream().limit(1))
//                .toList();
    }

    @Override
    public void close() throws IOException {
        in.close();
    }
}
